@@ -169,7 +169,7 @@ free_entries(GObject *object, gboolean should_signal)
 			IndicatorObjectEntry * entry;
 			entry = g_array_index(priv->entries, IndicatorObjectEntry *, 0);
 			g_array_remove_index(priv->entries, 0);
-			if (should_signal) {			
+			if (should_signal) {
 				g_signal_emit_by_name(object, WINDOW_MENU_SIGNAL_ENTRY_REMOVED, entry, TRUE);
 			}
 			entry_free(entry);
@@ -196,15 +196,11 @@ window_menu_dbusmenu_dispose (GObject *object)
 	}
 
 	if (priv->client != NULL) {
-		g_signal_handlers_disconnect_by_func(G_OBJECT(priv->client), G_CALLBACK(root_changed),    object);
-		g_signal_handlers_disconnect_by_func(G_OBJECT(priv->client), G_CALLBACK(event_status),    object);
-		g_signal_handlers_disconnect_by_func(G_OBJECT(priv->client), G_CALLBACK(item_activate),   object);
-		g_signal_handlers_disconnect_by_func(G_OBJECT(priv->client), G_CALLBACK(status_changed),  object);
-
+		g_signal_handlers_disconnect_by_data(priv->client, object);
 		g_object_unref(G_OBJECT(priv->client));
 		priv->client = NULL;
 	}
-	
+
 	if (priv->props != NULL) {
 		g_object_unref(G_OBJECT(priv->props));
 		priv->props = NULL;
@@ -513,8 +509,8 @@ remove_menuitem_signals (DbusmenuMenuitem * mi, gpointer user_data)
 {
 	g_signal_handlers_disconnect_by_func(G_OBJECT(mi), G_CALLBACK(menu_entry_realized), user_data);
 	g_signal_handlers_disconnect_by_func(G_OBJECT(mi), G_CALLBACK(menu_entry_realized_child_added), user_data);
-	g_signal_handlers_disconnect_matched (mi, G_SIGNAL_MATCH_FUNC, 0, 0, 0, menu_child_realized, NULL);
-	g_signal_handlers_disconnect_matched (mi, G_SIGNAL_MATCH_FUNC, 0, 0, 0, menu_prop_changed, NULL);
+	g_signal_handlers_disconnect_matched(mi, G_SIGNAL_MATCH_FUNC, 0, 0, 0, menu_child_realized, NULL);
+	g_signal_handlers_disconnect_matched(mi, G_SIGNAL_MATCH_FUNC, 0, 0, 0, menu_prop_changed, NULL);
 
 	return;
 }
@@ -531,9 +527,7 @@ root_changed (DbusmenuClient * client, DbusmenuMenuitem * new_root, gpointer use
 
 	if (priv->root != NULL) {
 		dbusmenu_menuitem_foreach(priv->root, remove_menuitem_signals, user_data);
-
-		g_signal_handlers_disconnect_by_func(G_OBJECT(priv->root), G_CALLBACK(menu_entry_added), user_data);
-		g_signal_handlers_disconnect_by_func(G_OBJECT(priv->root), G_CALLBACK(menu_entry_removed), user_data);
+		g_signal_handlers_disconnect_by_data(priv->root, user_data);
 		g_object_unref(priv->root);
 	}
 
@@ -670,8 +664,9 @@ menu_prop_changed (DbusmenuMenuitem * item, const gchar * property, GVariant * v
 		gtk_widget_set_sensitive(GTK_WIDGET(entry->label), g_variant_get_boolean(value));
 		wmentry->disabled = !g_variant_get_boolean(value);
 	} else if (!g_strcmp0(property, DBUSMENU_MENUITEM_PROP_LABEL)) {
-		gtk_label_set_text_with_mnemonic(entry->label, g_variant_get_string(value, NULL));
-
+		const gchar* str = g_variant_get_string(value, NULL);
+		if(str != NULL)
+			gtk_label_set_text_with_mnemonic(entry->label, str);
 		g_clear_pointer(&wmentry->vaccessible_desc, g_variant_unref);
 		entry->accessible_desc = g_variant_get_string(value, NULL);
 		wmentry->vaccessible_desc = g_variant_ref(value);
@@ -710,6 +705,7 @@ menu_child_realized (DbusmenuMenuitem * child, gpointer user_data)
 	WMEntry * wmentry = g_new0(WMEntry, 1);
 	wmentry->wm = wm;
 	IndicatorObjectEntry * entry = &wmentry->ioentry;
+	entry->parent_window = priv->windowid;
 
 	wmentry->mi = newentry;
 	g_object_ref(G_OBJECT(wmentry->mi));
